@@ -101,7 +101,7 @@ static inline struct bitmask* rt_numa_parse_cpustring(const char* s,
 	 * libnuma do not have this function.  A work around should be to run
 	 * your command with e.g. taskset -c 9-15 <command>
 	 */
-	return numa_parse_cpustring(s);
+	return numa_parse_cpustring((char *)s);
 #endif
 }
 
@@ -128,7 +128,7 @@ static int rt_numa_numa_node_of_cpu(int cpu)
 	int max_node, max_cpus;
 
 	max_node = numa_max_node();
-	max_cpus = sysconf(_SC_NPROCESSORS_CONF);
+	max_cpus = sysconf(_SC_NPROCESSORS_ONLN);
 
 	if (cpu > max_cpus) {
 		errno = EINVAL;
